@@ -43,9 +43,10 @@ class DeseqDataSet:
         DataFrame containing clinical information.
         Must be indexed by sample barcodes.
 
-    design_factor : str
-        Name of the column of clinical to be used as a design variable.
-        (default: 'high_grade').
+    design_factors : str or list[str]
+        Name of the columns of clinical to be used as design variables. If a list,
+        the last factor will be considered the variable of interest by default.
+        Only bi-level factors are supported. (default: 'condition').
 
     reference_level : str
         The factor to use as a reference. Must be one of the values taken by the design.
@@ -155,7 +156,7 @@ class DeseqDataSet:
         self,
         counts,
         clinical,
-        design_factor="high_grade",
+        design_factors="condition",
         reference_level=None,
         min_mu=0.5,
         min_disp=1e-8,
@@ -175,21 +176,28 @@ class DeseqDataSet:
         test_valid_counts(counts)
         self.counts = counts
 
-        # Import clinical data and convert design_column to string
-        self.clinical = clinical
-        if not self.counts.index.identical(self.clinical.index):
-            raise ValueError(
-                "The count matrix and clinical data should have the same index."
+        if set(self.counts.index) != set(clinical.index):
+            raise KeyError(
+                "The count matrix and clinical data "
+                "should contain the same sample indexes."
             )
-        self.design_factor = design_factor
-        if self.clinical[self.design_factor].isna().any():
-            raise ValueError("NaNs are not allowed in the design factor.")
-        self.clinical[self.design_factor] = self.clinical[self.design_factor].astype(str)
 
-        # Build the design matrix (splits the dataset in two cohorts)
+        # Import clinical data and convert design_column to string
+        self.clinical = clinical.loc[self.counts.index]
+        # Convert design_factors to list if a single string was provided.
+        self.design_factors = (
+            [design_factors] if isinstance(design_factors, str) else design_factors
+        )
+        if self.clinical[self.design_factors].isna().any().any():
+            raise ValueError("NaNs are not allowed in the design factors.")
+        self.clinical[self.design_factors] = self.clinical[self.design_factors].astype(
+            str
+        )
+
+        # Build the design matrix (splits the dataset in cohorts)
         self.design_matrix = build_design_matrix(
-            self.clinical,
-            design_factor,
+            clinical_df=self.clinical,
+            design_factors=self.design_factors,
             ref=reference_level,
             expanded=False,
             intercept=True,
@@ -248,6 +256,8 @@ class DeseqDataSet:
         Fits a negative binomial per gene, independently.
         """
 
+        num_vars = self.design_matrix.shape[-1]
+
         # Check that size factors are available. If not, compute them.
         if not hasattr(self, "size_factors"):
             self.fit_size_factors()
@@ -259,29 +269,54 @@ class DeseqDataSet:
         non_zero = ~(self.counts == 0).all()
         self.non_zero_genes = non_zero[non_zero].index
         counts_nonzero = self.counts.loc[:, self.non_zero_genes].values
-        m = counts_nonzero.shape[1]
+        num_genes = counts_nonzero.shape[1]
 
         # Convert design_matrix to numpy for speed
         X = self.design_matrix.values
         rough_disps = self._rough_dispersions.loc[self.non_zero_genes].values
         size_factors = self.size_factors.values
 
-        with parallel_backend("loky", inner_max_num_threads=1):
-            mu_hat_ = np.array(
-                Parallel(
+        # mu_hat is initialized differently depending on the number of different factor
+        # groups. If there are as many different factor combinations as design factors
+        # (intercept included), it is fitted with a linear model, otherwise it is fitted
+        # with a GLM (using rough dispersion estimates).
+        if len(self.design_matrix.value_counts()) == num_vars:
+            with parallel_backend("loky", inner_max_num_threads=1):
+                mu_hat_ = np.array(
+                    Parallel(
+                        n_jobs=self.n_processes,
+                        verbose=self.joblib_verbosity,
+                        batch_size=self.batch_size,
+                    )(
+                        delayed(fit_lin_mu)(
+                            counts=counts_nonzero[:, i],
+                            size_factors=size_factors,
+                            design_matrix=X,
+                            min_mu=self.min_mu,
+                        )
+                        for i in range(num_genes)
+                    )
+                )
+        else:
+            with parallel_backend("loky", inner_max_num_threads=1):
+                res = Parallel(
                     n_jobs=self.n_processes,
                     verbose=self.joblib_verbosity,
                     batch_size=self.batch_size,
                 )(
-                    delayed(fit_lin_mu)(
-                        counts_nonzero[:, i],
-                        size_factors,
-                        X,
-                        self.min_mu,
+                    delayed(irls_solver)(
+                        counts=counts_nonzero[:, i],
+                        size_factors=size_factors,
+                        design_matrix=X,
+                        disp=rough_disps[i],
+                        min_mu=self.min_mu,
+                        beta_tol=self.beta_tol,
                     )
-                    for i in range(m)
+                    for i in range(num_genes)
                 )
-            )
+
+                _, mu_hat_, _, _ = zip(*res)
+                mu_hat_ = np.array(mu_hat_)
 
         self._mu_hat = pd.DataFrame(
             mu_hat_.T,
@@ -294,14 +329,14 @@ class DeseqDataSet:
         with parallel_backend("loky", inner_max_num_threads=1):
             res = Parallel(n_jobs=self.n_processes, batch_size=self.batch_size)(
                 delayed(fit_alpha_mle)(
-                    counts_nonzero[:, i],
-                    X,
-                    mu_hat_[i, :],
-                    rough_disps[i],
-                    self.min_disp,
-                    self.max_disp,
+                    counts=counts_nonzero[:, i],
+                    design_matrix=X,
+                    mu=mu_hat_[i, :],
+                    alpha_hat=rough_disps[i],
+                    min_disp=self.min_disp,
+                    max_disp=self.max_disp,
                 )
-                for i in range(m)
+                for i in range(num_genes)
             )
         end = time.time()
         print(f"... done in {end - start:.2f} seconds.\n")
@@ -396,8 +431,8 @@ class DeseqDataSet:
             self.fit_dispersion_trend()
 
         # Exclude genes with all zeroes
-        m = len(self.non_zero_genes)
-        p = self.design_matrix.shape[1]
+        num_genes = len(self.non_zero_genes)
+        num_vars = self.design_matrix.shape[1]
 
         # Fit dispersions to the curve, and compute log residuals
         disp_residuals = np.log(self.genewise_dispersions) - np.log(
@@ -413,7 +448,7 @@ class DeseqDataSet:
             disp_residuals.loc[self.non_zero_genes][above_min_disp]
         ).median() ** 2 / norm.ppf(0.75)
         self.prior_disp_var = np.maximum(
-            self._squared_logres - polygamma(1, (m - p) / 2), 0.25
+            self._squared_logres - polygamma(1, (num_genes - num_vars) / 2), 0.25
         )
 
     def fit_MAP_dispersions(self):
@@ -428,7 +463,7 @@ class DeseqDataSet:
 
         # Exclude genes with all zeroes
         counts_nonzero = self.counts.loc[:, self.non_zero_genes].values
-        m = counts_nonzero.shape[1]
+        num_genes = counts_nonzero.shape[1]
 
         X = self.design_matrix.values
         mu_hat = self._mu_hat.values
@@ -443,17 +478,17 @@ class DeseqDataSet:
                 batch_size=self.batch_size,
             )(
                 delayed(fit_alpha_mle)(
-                    counts_nonzero[:, i],
-                    X,
-                    mu_hat[:, i],
-                    fit_disps[i],
-                    self.min_disp,
-                    self.max_disp,
-                    self.prior_disp_var,
-                    True,
-                    True,
+                    counts=counts_nonzero[:, i],
+                    design_matrix=X,
+                    mu=mu_hat[:, i],
+                    alpha_hat=fit_disps[i],
+                    min_disp=self.min_disp,
+                    max_disp=self.max_disp,
+                    prior_disp_var=self.prior_disp_var,
+                    cr_reg=True,
+                    prior_reg=True,
                 )
-                for i in range(m)
+                for i in range(num_genes)
             )
         end = time.time()
         print(f"... done in {end-start:.2f} seconds.\n")
@@ -489,7 +524,7 @@ class DeseqDataSet:
 
         # Exclude genes with all zeroes
         counts_nonzero = self.counts.loc[:, self.non_zero_genes].values
-        m = counts_nonzero.shape[1]
+        num_genes = counts_nonzero.shape[1]
 
         X = self.design_matrix.values
         size_factors = self.size_factors.values
@@ -504,14 +539,14 @@ class DeseqDataSet:
                 batch_size=self.batch_size,
             )(
                 delayed(irls_solver)(
-                    counts_nonzero[:, i],
-                    size_factors,
-                    X,
-                    disps[i],
-                    self.min_mu,
-                    self.beta_tol,
+                    counts=counts_nonzero[:, i],
+                    size_factors=size_factors,
+                    design_matrix=X,
+                    disp=disps[i],
+                    min_mu=self.min_mu,
+                    beta_tol=self.beta_tol,
                 )
-                for i in range(m)
+                for i in range(num_genes)
             )
         end = time.time()
         print(f"... done in {end-start:.2f} seconds.\n")
@@ -558,7 +593,7 @@ class DeseqDataSet:
         if not hasattr(self, "dispersions"):
             self.fit_MAP_dispersions()
 
-        p = self.design_matrix.shape[1]
+        num_vars = self.design_matrix.shape[1]
         # Keep only non-zero genes
         normed_counts = self.counts.loc[:, self.non_zero_genes].div(self.size_factors, 0)
         dispersions = robust_method_of_moments_disp(normed_counts, self.design_matrix)
@@ -568,7 +603,7 @@ class DeseqDataSet:
         ) ** 2 / V
         self.cooks = (
             squared_pearson_res
-            / p
+            / num_vars
             * (self._hat_diagonals / (1 - self._hat_diagonals) ** 2)
         ).T
 
@@ -609,7 +644,7 @@ class DeseqDataSet:
         if not hasattr(self, "cooks"):
             self.calculate_cooks()
 
-        m, p = self.design_matrix.shape
+        num_samples, num_vars = self.design_matrix.shape
 
         # Check whether cohorts have enough samples to allow refitting
         n_or_more = (
@@ -622,7 +657,7 @@ class DeseqDataSet:
         self.replaceable.index = self.design_matrix.index
 
         # Get positions of counts with cooks above threshold
-        cooks_cutoff = f.ppf(0.99, p, m - p)
+        cooks_cutoff = f.ppf(0.99, num_vars, num_samples - num_vars)
         idx = (self.cooks > cooks_cutoff).T
         self.replaced = idx.any(axis=0)
 
@@ -671,7 +706,7 @@ class DeseqDataSet:
         sub_dds = DeseqDataSet(
             counts=self.counts_to_refit,
             clinical=self.clinical,
-            design_factor=self.design_factor,
+            design_factors=self.design_factors,
             min_mu=self.min_mu,
             min_disp=self.min_disp,
             max_disp=self.max_disp,
